#include "MultiGridSolver.h"
#include "FiniteElemSpace.h"
#include "DOFVector.h"
#include "DOFMatrix.h"
#include "Traverse.h"
#include "MacroWriter.h"
#include "ValueWriter.h"
#include "SparseVector.h" 
#include "ElInfo.h"
#include "Parameters.h"
#include "SystemVector.h"
#include "GSSmoother.h"
#include <algorithm>
#include "InterpolRestrictMatrix.h"
#include "LagrangeInterpolRestrict.h"
#include "DirichletBC.h"

namespace AMDiS {

  MultiGridSolverScal::MultiGridSolverScal(const ::std::string &name)
    : MultiGridSolverBase<DOFMatrix, DOFVector<double> >(name), 
      sparseResidual_(NULL),
      denseResidual_(NULL),
      lastRes_(1.0),
      iteration_(0),
      coarseSmoothingSteps_(2),
      minLevelGap_(0),
      maxMGLevels_(100),
      interpolMatrix_(NULL),
      restrictMatrix_(NULL),
      galerkin_(0)
  {
    FUNCNAME("MultiGridSolverScal::MultiGridSolverScal()");

    GET_PARAMETER(0, name_ + "->coarse level smoothing steps", "%d", &coarseSmoothingSteps_);
    GET_PARAMETER(0, name_ + "->min level gap", "%d", &minLevelGap_);
    GET_PARAMETER(0, name_ + "->max mg levels", "%d", &maxMGLevels_);
    GET_PARAMETER(0, name_ + "->use galerkin operator", "%d", &galerkin_);

    // Create smoother
    ::std::string smootherType("gs");
    GET_PARAMETER(0, name_ + "->smoother", &smootherType);
    SmootherCreator<DOFMatrix, SparseVector<double>, ::std::set<DegreeOfFreedom> > *smootherCreator =
      dynamic_cast<SmootherCreator<DOFMatrix, SparseVector<double>, ::std::set<DegreeOfFreedom> >*>(
												    CreatorMap<SmootherBase<DOFMatrix, SparseVector<double>, ::std::set<DegreeOfFreedom> > >::getCreator(smootherType));

    if (smootherCreator && !smootherCreator->isNullCreator()) {
      smootherCreator->setName(name_ + "->smoother");
      smoother_ = smootherCreator->create();
    } else {
      smoother_ = NULL;
      ERROR_EXIT("no smoother specified\n");
    }
  }

  void MultiGridSolverScal::initMaxLevel()
  {
    FiniteElemSpace *feSpace = 
      const_cast<FiniteElemSpace*>(systemMatrix_->getFESpace());

    TraverseStack st;
    ElInfo *elInfo = st.traverseFirst(feSpace->getMesh(), -1, 
				      Mesh::CALL_LEAF_EL);
    maxLevel_ = -1;
    while (elInfo) {
      maxLevel_ = max(elInfo->getLevel(), maxLevel_);
      elInfo = st.traverseNext(elInfo);   
    }
  }

  void MultiGridSolverScal::initMultiGrid(bool initMG)
  {
    FUNCNAME("MultiGridSolverScal::initMultiGrid()");

    iteration_ = 0;

    lastRes_ = 1.0;

    if (initMG) {
      FiniteElemSpace *feSpace = 
	const_cast<FiniteElemSpace*>(systemMatrix_->getFESpace());

      TEST_EXIT(!galerkin_ || feSpace->getBasisFcts()->getDegree() == 1)
	("galerkin coarse grid operator only for linear finite elements\n");

      Mesh *mesh = feSpace->getMesh();

      int dim = mesh->getDim();
      int degree = feSpace->getBasisFcts()->getDegree();
      
      // Get the interpolation and restriction matrices that correspond to
      // the element dim and to the degree of the basis functions.
      interpolMatrix_ = interpolMatrices[dim - 1][degree - 1];
      restrictMatrix_ = restrictMatrices[dim - 1][degree - 1];

      sparseResidual_ = NEW SparseVector<double>(feSpace, "sparse residual");
      denseResidual_ = NEW DOFVector<double>(feSpace, "dense residual");

      int numLevels = maxLevel_ + 1;

      isMGLevel_.resize(numLevels, false);
      isMGLevel_[minLevel_] = true;
      isMGLevel_[maxLevel_] = true;
      if (numLevels > 2) { 
	double step = (static_cast<double>(maxMGLevels_) - 2.0) / (numLevels - 2.0);
	double sum = 0.0;
	int gap = 0;
	for (int i = maxLevel_ - 1; i > minLevel_; i--) {
	  sum += step;
	  //MSG("%d %f\n", i, sum);
	  ++gap;
	  if(gap > minLevelGap_ && sum >= 1.0) {
	    gap = 0;
	    isMGLevel_[i] = true;
	    sum -= 1.0;	
	  } else {
	    MSG("level %d skipped\n", i);
	  }
	}
      }

      int error = mesh->queryCoarseDOFs();
      TEST_EXIT(error)("preserve coarse dofs not set in mesh\n");

      const BasisFunction *basFcts = feSpace->getBasisFcts();
      int numDOFs = basFcts->getNumber();

      // Allocate memory for the multigrid levels.
      levels_.resize(numLevels);

      // Fill all multigrid levels with initial data.
      for (int i = 0; i < numLevels; i++) {
	if (isMGLevel_[i]) {
	  levels_[i].solution_ = NEW SparseVector<double>(feSpace, "sparse level solution");
	  levels_[i].oldSolution_ = NEW SparseVector<double>("old level solution", levels_[i].solution_);
	  levels_[i].rhs_ = NEW SparseVector<double>("sparse level rhs", levels_[i].solution_);
	  levels_[i].matrix_ = NEW DOFMatrix(feSpace, feSpace, "level matrix");
	  levels_[i].matrix_->setBoundaryManager(systemMatrix_->getBoundaryManager());
	}
      }
    
      // Here, for each DOF we save the highest mesh level on which it occurs.
      DOFVector<int> maxDOFLevels(feSpace, "max dof levels");
      maxDOFLevels.set(minLevel_);
    
      DegreeOfFreedom *dofs = GET_MEMORY(DegreeOfFreedom, numDOFs);
      Element *element;
      DOFAdmin *admin = feSpace->getAdmin();
      int level;

      // Iterators for all operators of the system matrix.
      ::std::vector<Operator*>::iterator it;
      ::std::vector<Operator*>::iterator opBegin = systemMatrix_->getOperatorsBegin();
      ::std::vector<Operator*>::iterator opEnd = systemMatrix_->getOperatorsEnd();

      ::std::vector<double*>::iterator factorIt;
      ::std::vector<double*>::iterator factorBegin = systemMatrix_->getOperatorFactorBegin();
      ElementMatrix *elementMatrix = NULL; 
      const BoundaryType *bound;
    
      TraverseStack stack;
    
      // Fill level elements and max-dof-level for each DOF.
      ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_EVERY_EL_PREORDER);
      while (elInfo) {
	element = elInfo->getElement();
	level = elInfo->getLevel();
	levels_[level].elements_.push_back(element);
	if (dim == 3) {
	  levels_[level].elementTypes_.push_back(dynamic_cast<ElInfo3d*>(elInfo)->getType());
	}
	basFcts->getLocalIndices(element, admin, dofs);
	for (int i = 0; i < numDOFs; i++) {
	  maxDOFLevels[dofs[i]] = max(maxDOFLevels[dofs[i]], mgLevel(level));
	}
	elInfo = stack.traverseNext(elInfo);
      }
    
    
      // Assemble level matrices and create DOF sets.
      elInfo = stack.traverseFirst(mesh, -1, 
				   Mesh::CALL_EVERY_EL_PREORDER |
				   Mesh::FILL_BOUND |
				   Mesh::FILL_DET |
				   Mesh::FILL_GRD_LAMBDA |
				   Mesh::FILL_COORDS |
				   Mesh::FILL_NEIGH);   
      while (elInfo) {
	element = elInfo->getElement();
	level = elInfo->getLevel();
	basFcts->getLocalIndices(element, admin, dofs);
	bound = basFcts->getBound(elInfo, NULL);
	elementMatrix = (*opBegin)->getAssembler()->initElementMatrix(elementMatrix, elInfo);
      
	bool levelUsed = isMGLevel_[level];
	
	// If the element is in a level that is a multigrid level, add all element
	// matrices of this element to the level matrix.
	if ((levelUsed) && (!galerkin_)) {
	  for (it = opBegin, factorIt = factorBegin;	
	       it != opEnd; 
	       ++it, ++factorIt) 
	    {
	      (*it)->getElementMatrix(elInfo, 
				      elementMatrix, 
				      *factorIt ? **factorIt : 1.0);
	    }
	  
	  levels_[level].matrix_->addElementMatrix(1.0, *elementMatrix, bound);	
	}

	if (element->isLeaf()) {
	  // If the element is a leaf element, but this level is not a multigrid level,
	  // the element matrices are added to the next possible multigrid level.
	  if (!levelUsed) {
	    level = mgLevel(level);
	    if(!galerkin_) {
	      for(it = opBegin, factorIt = factorBegin;	
		  it != opEnd; 
		  ++it, ++factorIt) 
		{
		  (*it)->getElementMatrix(elInfo, 
					  elementMatrix, 
					  *factorIt ? **factorIt : 1.0);
		}
	      levels_[level].matrix_->addElementMatrix(1.0, *elementMatrix, bound);	   
	    }
	  }
	
	  // Add nearest neighbour element matrices and dofs
	  int maxDOFLevel = -1;
	  for (int i = 0; i < dim + 1; i++) {
	    maxDOFLevel = max(maxDOFLevel, maxDOFLevels[dofs[i]]);
	  }
	
	  // There is a DOF of the element that is included in finer levels.
	  if (maxDOFLevel > level) {
	    if (!galerkin_) {
	      for (int j = level + 1; j < maxDOFLevel + 1; j++) {
		if (isMGLevel_[j]) {
		  levels_[j].matrix_->addElementMatrix(1.0, *elementMatrix, bound);
		}
	      }
	    }
	    for (int j = 0; j < numDOFs; j++) {
	      for (int k = level + 1; k < maxDOFLevel + 1; k++) {
		if (isMGLevel_[k]) {
		  if (maxDOFLevels[dofs[j]] < k) {
		    levels_[k].nearestNeighbourDOFs_.insert(dofs[j]);
		  }
		}
	      }
	    }
	  }
	
	  for (int i = 0; i < numDOFs; i++) {
	    // If the element is a leaf element and the maximum level of one of
	    // its DOFs is equal to the current level, the DOF is not included
	    // in finer levels.
	    if (maxDOFLevels[dofs[i]] == level) {
	      levels_[level].coarseDOFs_.insert(dofs[i]);
	    }
	  }
	} else { 
	  // element is not leaf
	  if ((level >= minLevel_) && (levelUsed)) {
	    // The element is not a leaf element, therefore its DOFs are also
	    // included in finer levels.
	    for (int i = 0; i < numDOFs; i++) {
	      levels_[level].fineDOFs_.insert(dofs[i]);
	    }
	  }
	} // if-else element is leaf
	elInfo = stack.traverseNext(elInfo);
      } // while elInfo
    
      if (!galerkin_) {
	// Fill boundary conditions
	if (systemMatrix_->getBoundaryManager()) {
	  systemMatrix_->getBoundaryManager()->initMatrix(systemMatrix_);
	}
      
	TraverseStack stack;
	ElInfo *elInfo = stack.traverseFirst(mesh, -1, 
					     Mesh::CALL_LEAF_EL | 
					     Mesh::FILL_BOUND |
					     Mesh::FILL_COORDS |
					     Mesh::FILL_DET |
					     Mesh::FILL_GRD_LAMBDA |
					     Mesh::FILL_NEIGH);
	while (elInfo) {
	  if (systemMatrix_->getBoundaryManager()) {
	    systemMatrix_->getBoundaryManager()->fillBoundaryConditions(elInfo,
									systemMatrix_);
	  }

	  elInfo = stack.traverseNext(elInfo);
	}
      
	if (systemMatrix_->getBoundaryManager()) {
	  systemMatrix_->getBoundaryManager()->exitMatrix(systemMatrix_);
	}
      } else { 
	// Galerkin operator

	DegreeOfFreedom *child0DOFs = GET_MEMORY(DegreeOfFreedom, dim + 1);
	DegreeOfFreedom *parentDOFs = GET_MEMORY(DegreeOfFreedom, dim + 1);
      
	Vector<DegreeOfFreedom> fineDOFs(dim + 2);
	Matrix<double> a(dim + 2, dim + 2);
 
	levels_[maxLevel_].matrix_->copy(*systemMatrix_);

	DOFMatrix tmp1(feSpace, feSpace, "tmp1");
	DOFMatrix tmp2(feSpace, feSpace, "tmp2");

	DOFMatrix *coarseMatrix = NULL;
	DOFMatrix *fineMatrix = levels_[maxLevel_].matrix_;

	for (level = maxLevel_ - 1; level >= minLevel_; level--) {
	  ::std::map< ::std::pair<int, int>, bool> visited;

	  coarseMatrix =
	    isMGLevel_[level] ?
	    levels_[level].matrix_ :
	    (fineMatrix == &tmp1 ? &tmp2 : &tmp1);

	  coarseMatrix->clear();

	  elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_EVERY_EL_PREORDER);
	  while(elInfo) {
	    Element *element = elInfo->getElement();

	    basFcts->getLocalIndices(element, admin, parentDOFs);

	    for (int i = 0; i < dim + 1; i++) {
	      fineDOFs[i] = parentDOFs[i];
	    }

	    if(elInfo->getLevel() == level && !element->isLeaf()) {

	      basFcts->getLocalIndices(element->getChild(0), admin, child0DOFs);

	      fineDOFs[dim + 1] = child0DOFs[dim];

	      double *ptr;

	      for (int i = 0; i < dim + 2; i++) {
		for (int j = 0; j < dim + 2; j++) {
		  ptr = fineMatrix->hasSparseDOFEntry(fineDOFs[i], fineDOFs[j]);
		  a[i][j] = ptr ? *ptr : 0.0;
		}
	      }

	      for (int i = 0; i < dim + 1; i++) {
		for (int j = 0; j < dim + 1; j++) {
		  if(visited[::std::pair<int, int>(fineDOFs[i], fineDOFs[j])] == false)
		    coarseMatrix->addSparseDOFEntry(1.0, 
						    parentDOFs[i],
						    parentDOFs[j],
						    0.0,
						    false);		
		}
	      }

	      for (int i = 0; i < dim + 1; i++) {
		for (int j = 0; j < dim + 1; j++) {
		  if(visited[::std::pair<int, int>(fineDOFs[i], fineDOFs[j])] == false)
		    coarseMatrix->addSparseDOFEntry(1.0, 
						    parentDOFs[i],
						    parentDOFs[j],
						    a[i][j],
						    true);
		}
	      }

	      for (int i = 0; i < dim + 1; i++) {
		for (int j = 0; j < 2; j++) {
		  if(visited[::std::pair<int, int>(fineDOFs[i], fineDOFs[dim + 1])] == false) {
		    coarseMatrix->addSparseDOFEntry(0.5, 
						    parentDOFs[i],
						    parentDOFs[j],
						    a[i][dim + 1],
						    true);	      

		    coarseMatrix->addSparseDOFEntry(0.5, 
						    parentDOFs[j],
						    parentDOFs[i],
						    a[dim + 1][i],
						    true);
		  }	      
		}
	      }

	      if (visited[::std::pair<int, int>(fineDOFs[dim + 1], fineDOFs[dim + 1])] == false) {
		for (int i = 0; i < 2; i++) {
		  for (int j = 0; j < 2; j++) {
		    coarseMatrix->addSparseDOFEntry(0.25, 
						    parentDOFs[i],
						    parentDOFs[j],
						    a[dim + 1][dim + 1],
						    true);	      
		  }
		}
	      }

	      for (int i = 0; i < dim + 2; i++) {
		for (int j = 0; j < dim + 2; j++) {
		  visited[::std::pair<int, int>(fineDOFs[i], fineDOFs[j])] = true;
		  visited[::std::pair<int, int>(fineDOFs[j], fineDOFs[i])] = true;
		}
	      }
	    } 
	    if (element->isLeaf() && elInfo->getLevel() <= level) {
	      double *ptr;

	      for (int i = 0; i < dim + 1; i++) {
		for (int j = 0; j < dim + 1; j++) {
		  ptr = fineMatrix->hasSparseDOFEntry(fineDOFs[i], fineDOFs[j]);
		  a[i][j] = ptr ? *ptr : 0.0;
		}
	      }

	      for (int i = 0; i < dim + 1; i++) {
		for (int j = 0; j < dim + 1; j++) {
		  if (!visited[::std::pair<int, int>(parentDOFs[i], parentDOFs[j])]) {
		    coarseMatrix->addSparseDOFEntry(1.0, 
						    parentDOFs[i],
						    parentDOFs[j],
						    a[i][j],
						    false);	      		
		  }
		}
	      }	  
	    }
	    elInfo = stack.traverseNext(elInfo);
	  }
	  fineMatrix = coarseMatrix;
	} // for level

	FREE_MEMORY(child0DOFs, DegreeOfFreedom, dim + 1);
	FREE_MEMORY(parentDOFs, DegreeOfFreedom, dim + 1);
      } // if-else galerkin_

      FREE_MEMORY(dofs, DegreeOfFreedom, numDOFs);

      // Copy in each level the to sets coarseDOFs and fineDOFs to one
      // set levelDOFs containing all DOFs of the level.
      for (int i = minLevel_; i < numLevels; i++) {
	if (isMGLevel_[i]) {
	  levels_[i].levelDOFs_.insert(levels_[i].coarseDOFs_.begin(),
				       levels_[i].coarseDOFs_.end());
	  levels_[i].levelDOFs_.insert(levels_[i].fineDOFs_.begin(),
				       levels_[i].fineDOFs_.end());
	}
      }
    } // if initMG

    // Compute start residual
    mv(NoTranspose, *systemMatrix_, *solution_, *denseResidual_);
    axpy(-1.0, *rhs_, *denseResidual_);

    // Compute 2-norm of the residual
    double res = denseResidual_->nrm2();
    MSG("residual: %e\n", res);
    lastRes_ = res;

    for (int i = 0; i <= maxLevel_; i++) {
      if (isMGLevel_[i]) {
	levels_[i].solution_->copyFromDOFVector(*solution_, 
						levels_[i].coarseDOFs_);
	levels_[i].solution_->copyFromDOFVector(*solution_, 
						levels_[i].nearestNeighbourDOFs_);
      }
    }

    // Copy the rhs of the equation to the finest multigrid level
    // restricted to its DOFs.
    levels_[maxLevel_].rhs_->copyFromDOFVector(*rhs_, 
					       levels_[maxLevel_].coarseDOFs_);
  }
  

  void MultiGridSolverScal::exitMultiGrid()
  {
    DELETE sparseResidual_;
    DELETE denseResidual_;

    // Free memory
    for (int i = 0; i < maxLevel_ + 1; i++) {
      if (isMGLevel_[i]) {
	DELETE levels_[i].matrix_;
	DELETE levels_[i].rhs_;
	DELETE levels_[i].oldSolution_;
	DELETE levels_[i].solution_;
      }
    }

    levels_.resize(0);
    isMGLevel_.resize(0);
  }


  void MultiGridSolverScal::smooth(int level, int steps)
  { 
    smoother_->smooth(levels_[level].matrix_, 
		      levels_[level].solution_,
		      levels_[level].rhs_,
		      steps,
		      levels_[level].levelDOFs_);
    //   DOFVector<double> sol(systemMatrix_->getFESpace(), "solution");
    //   levels_[level].solution_->copyToDOFVector(sol);
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "sol.mesh", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&sol, "sol.dat", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   MSG("smoothed %d\n", level);
    //   WAIT;
  }


  void MultiGridSolverScal::prolongate(int level, int coarseLevel)
  {
    for (int i = coarseLevel; i < level; i++) {
      interpolMatrix_->mv(sparseResidual_, 
			  sparseResidual_, 
			  i);
    }
    //   DOFVector<double> cor(systemMatrix_->getFESpace(), "correction");
    //   sparseResidual_->copyToDOFVector(cor);
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "cor.mesh", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&cor, "cor.dat", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   MSG("prolongated %d\n", level);
    //   WAIT;
  }


  void MultiGridSolverScal::restrict(int level, int coarseLevel)
  {
    //   MSG("%d\n", level);
    //   if(isMGLevel_[level]) levels_[level].rhs_->print();

    // Copy the old solution to the coarser grid. 
    levels_[coarseLevel].oldSolution_->copy(*(levels_[level].solution_),
					    levels_[level].levelDOFs_);
    levels_[coarseLevel].oldSolution_->copy(*(levels_[level].solution_),
					    levels_[level].nearestNeighbourDOFs_);

    ::std::vector<Element*>::iterator elIt, elBegin, elEnd;
    ::std::vector<int>::iterator typeIt;

    for (int i = level; i > coarseLevel; i--) {
      elBegin = levels_[i-1].elements_.begin();
      elEnd = levels_[i-1].elements_.end();
      typeIt = levels_[i-1].elementTypes_.begin();

      // Restrict the old solution to the DOFs of the coarser gird.
      for (elIt = elBegin; elIt != elEnd; ++elIt) {
	Element *element = *elIt;
	if (!element->isLeaf()) {
	  RCNeighbourList list(1);
	  list.setElement(0, element);
	  if (element->getMesh()->getDim() == 3) {
	    list.setType(0, levels_[i-1].elementTypes_[*typeIt]);
	    ++typeIt;
	  }
	  levels_[coarseLevel].oldSolution_->coarseRestrict(list, 1);
	}
      }

      // Restrict the residual
      restrictMatrix_->mv(sparseResidual_, 
			  sparseResidual_, 
			  i - 1);

      // Alternative: use restriction matrix to restrict the old solution.
      //     restrictMatrix_->mv(levels_[coarseLevel].oldSolution_, 
      // 			levels_[coarseLevel].oldSolution_,
      // 			i-1);
    }


    levels_[coarseLevel].solution_->copy(*(levels_[coarseLevel].oldSolution_),
					 levels_[coarseLevel].fineDOFs_);

    //   DOFVector<double> res(systemMatrix_->getFESpace(), "residual");
    //   sparseResidual_->copyToDOFVector(res);
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "res.mesh", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&res, "res.dat", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   DOFVector<double> sol(systemMatrix_->getFESpace(), "solution");
    //   levels_[coarseLevel].oldSolution_->copyToDOFVector(sol);
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "sol.mesh", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&sol, "sol.dat", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   MSG("residual and solution restricted %d\n", level);
    //   WAIT;
  }


  void MultiGridSolverScal::computeResidual(int level, int coarseLevel)
  {
    mv(*(levels_[level].matrix_), 
       *(levels_[level].solution_), 
       *sparseResidual_, 
       levels_[level].levelDOFs_);

    axpy(-1.0, 
	 *(levels_[level].rhs_), 
	 *sparseResidual_, 
	 levels_[level].levelDOFs_);

    //   DOFVector<double> res(systemMatrix_->getFESpace(), "residual");
    //   sparseResidual_->copyToDOFVector(res);
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "res.mesh", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&res, "res.dat", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   MSG("residual computed %d\n", level);
    //   WAIT;
  }


  void MultiGridSolverScal::computeRHS(int level, int coarseLevel)
  {
    // First compute rhs without considering boundary conditions.
    mv(*(levels_[coarseLevel].matrix_), 
       *(levels_[coarseLevel].oldSolution_),
       *(levels_[coarseLevel].rhs_),
       levels_[coarseLevel].fineDOFs_);

    axpy(-1.0, 
	 *sparseResidual_, 
	 *(levels_[coarseLevel].rhs_), 
	 levels_[coarseLevel].fineDOFs_);

    levels_[coarseLevel].rhs_->copyFromDOFVector(*rhs_, 
						 levels_[coarseLevel].coarseDOFs_);


    // Then add boundary conditions to the solution.
    if (!galerkin_) {
      Mesh *mesh = rhs_->getFESpace()->getMesh();
  
      if(rhs_->getBoundaryManager()) {
	rhs_->getBoundaryManager()->initVector(levels_[coarseLevel].rhs_);
      }

      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(mesh, -1, 
					   Mesh::CALL_EVERY_EL_PREORDER |
					   Mesh::FILL_BOUND |
					   Mesh::FILL_DET |
					   Mesh::FILL_GRD_LAMBDA |
					   Mesh::FILL_COORDS);

      while (elInfo) {
	if (rhs_->getBoundaryManager())
	  rhs_->getBoundaryManager()->fillBoundaryConditions(elInfo, 
							     levels_[coarseLevel].rhs_);
	elInfo = stack.traverseNext(elInfo);
      }

      if (rhs_->getBoundaryManager()) {
	rhs_->getBoundaryManager()->exitVector(levels_[coarseLevel].rhs_);
      }
    }

    //   MSG("%d\n", level);
    //   levels_[level].matrix_->print();
    //   levels_[level].solution_->print();
    //   levels_[level].rhs_->print();
    //   MSG("%d\n", coarseLevel);
    //   levels_[coarseLevel].matrix_->print();
    //   levels_[coarseLevel].solution_->print();
    //   levels_[coarseLevel].rhs_->print();
  
    //   DOFVector<double> rhs(systemMatrix_->getFESpace(), "rhs");
    //   levels_[coarseLevel].rhs_->copyToDOFVector(rhs);
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "rhs.mesh", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&rhs, "rhs.dat", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "realrhs.mesh");
    //   ValueWriter::writeValues(rhs_, "realrhs.dat");
    //   MSG("rhs computed %d\n", level);
    //   WAIT;
  }


  void MultiGridSolverScal::computeCorrection(int level, int coarseLevel)
  {
    sparseResidual_->copy(*(levels_[coarseLevel].solution_), 
			  levels_[coarseLevel].fineDOFs_);
    axpy(-1.0, 
	 *(levels_[coarseLevel].oldSolution_), 
	 *sparseResidual_, 
	 levels_[coarseLevel].fineDOFs_);

    //   DOFVector<double> cor(systemMatrix_->getFESpace(), "correction");
    //   sparseResidual_->copyToDOFVector(cor);
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "cor.mesh", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&cor, "cor.dat", 0.0, coarseLevel, Mesh::CALL_EL_LEVEL);
    //   MSG("correction computed %d\n", level);
    //   WAIT;
  } 


  void MultiGridSolverScal::fixSolution(int level, int coarseLevel)
  {
    levels_[level].solution_->copy(*(levels_[coarseLevel].solution_), 
				   levels_[level].nearestNeighbourDOFs_);

    axpy(1.0, 
	 *sparseResidual_, 
	 *(levels_[level].solution_), 
	 levels_[level].levelDOFs_);

    //   DOFVector<double> sol(systemMatrix_->getFESpace(), "solution");
    //   levels_[level].solution_->copyToDOFVector(sol);
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "sol.mesh", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&sol, "sol.dat", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   MSG("solution fixed %d\n", level);
    //   WAIT;
  }


  void MultiGridSolverScal::exactSolve(int level)
  {
    smooth(level, coarseSmoothingSteps_);
  }


  bool MultiGridSolverScal::toleranceReached(double tol)
  { 
    FUNCNAME("MultiGridSolver::toleranceReached()");

    for (int level = minLevel_; level < maxLevel_ + 1; level++) {
      if (isMGLevel_[level]) {
	levels_[level].solution_->copyToDOFVector(*solution_, 
						  levels_[level].coarseDOFs_);
      }
    }

    mv(NoTranspose, *systemMatrix_, *solution_, *denseResidual_);
    axpy(-1.0, *rhs_, *denseResidual_);

    ++iteration_;

    double res = denseResidual_->nrm2();
    MSG("iteration %d - residual: %e (res / lastRes : %e)\n", 
	iteration_, res, res/lastRes_);

    lastRes_ = res;

    //   levelMatrices_[maxLevel_]->print();

    //  systemMatrix_->print();

    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "sol_iter.mesh");
    //   ValueWriter::writeValues(solution_, "sol_iter.dat");
    //   MacroWriter::writeMacro(systemMatrix_->getFESpace(), "res_iter.mesh");
    //   ValueWriter::writeValues(denseResidual_, "res_iter.dat");
    //   WAIT;

    return (res < tol);
  }


  void MultiGridSolverScal::initFullMultiGrid()
  {
    int coarseLevel = -1;
  
    for (int i = maxLevel_; i > minLevel_; i--) {
      if (isMGLevel_[i]) {
	coarseLevel = coarserLevel(i);
	levels_[coarseLevel].rhs_->copy(*(levels_[i].rhs_), 
					levels_[coarseLevel].fineDOFs_);
      }

      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(sparseResidual_->getFESpace()->getMesh(),
					   -1, Mesh::CALL_EVERY_EL_PREORDER); 
      while (elInfo) {
	Element *element = elInfo->getElement();
	if (elInfo->getLevel() == (i-1) && !element->isLeaf()) {
	  RCNeighbourList list(1);
	  list.setElement(0, element);
	  if (element->getMesh()->getDim() == 3) {
	    list.setType(0, dynamic_cast<ElInfo3d*>(elInfo)->getType());
	  }
	  levels_[coarseLevel].rhs_->coarseRestrict(list, 1);
	}
	elInfo = stack.traverseNext(elInfo);
      }
    }
  }


  void MultiGridSolverScal::fmgInterpol(int level)
  {
    int coarseLevel = level-1;

    levels_[level].solution_->copy(*(levels_[coarseLevel].solution_), 
				   levels_[coarseLevel].fineDOFs_);

    for (int i = coarseLevel; i < level; i++) {

      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(sparseResidual_->getFESpace()->getMesh(),
					   -1, Mesh::CALL_EVERY_EL_PREORDER); 
      while(elInfo) {
	Element *element = elInfo->getElement();
	if(elInfo->getLevel() == (i) && !element->isLeaf()) {
	  RCNeighbourList list(1);
	  list.setElement(0, element);
	  if(element->getMesh()->getDim() == 3) {
	    list.setType(0, dynamic_cast<ElInfo3d*>(elInfo)->getType());
	  }
	  levels_[level].solution_->refineInterpol(list, 1);
	}
	elInfo = stack.traverseNext(elInfo);
      }
    }
  }

  // ===============================================================================
  // ===============================================================================
  // ===============================================================================
  // ===============================================================================

  MultiGridSolverVec::MultiGridSolverVec(const ::std::string &name)
    : MultiGridSolverBase<Matrix<DOFMatrix*>, SystemVector>(name),
      sparseResidual_(NULL),
      denseResidual_(NULL),
      lastRes_(1.0),
      iteration_(0),
      coarseSmoothingSteps_(3),
      minLevelGap_(0),
      maxMGLevels_(100),
      numComponents_(0),
      boxSmoothing_(false),
      boxDOFs_(0),
      galerkin_(0)
  {
    FUNCNAME("MultiGridSolverVec::MultiGridSolverVec()");

    GET_PARAMETER(0, name_ + "->coarse level smoothing steps", "%d", 
		  &coarseSmoothingSteps_);
    GET_PARAMETER(0, name_ + "->min level gap", "%d", &minLevelGap_);
    GET_PARAMETER(0, name_ + "->max mg levels", "%d", &maxMGLevels_);
    GET_PARAMETER(0, name_ + "->use galerkin operator", "%d", &galerkin_);

    // Create smoother
    ::std::string smootherType("gs");
    GET_PARAMETER(0, name_ + "->smoother", &smootherType);
    if(smootherType == "box") {
      boxSmoothing_ = true;
    }
    SmootherCreator<Matrix<DOFMatrix*>, Vector<SparseVector<double>*>, Vector< ::std::set<DegreeOfFreedom>*> > *smootherCreator =
      dynamic_cast<SmootherCreator<Matrix<DOFMatrix*>, Vector<SparseVector<double>*>, Vector< ::std::set<DegreeOfFreedom>*> >*>(
																CreatorMap<SmootherBase<Matrix<DOFMatrix*>, Vector<SparseVector<double>*>, Vector< ::std::set<DegreeOfFreedom>*> > >::getCreator(smootherType)
																);
    if(smootherCreator && !smootherCreator->isNullCreator()) {
      smootherCreator->setName(name_ + "->smoother");
      smoother_ = smootherCreator->create();
    } else {
      smoother_ = NULL;
      ERROR_EXIT("no smoother specified\n");
    }
  }

  void MultiGridSolverVec::initMaxLevel()
  {
    int comp, numComps = solution_->getNumVectors();
    ::std::set<Mesh*> meshes;
    for(comp = 0; comp < numComps; comp++) {
      Mesh *mesh = solution_->getDOFVector(comp)->getFESpace()->getMesh();
      if(find(meshes.begin(), meshes.end(), mesh) == meshes.end()) {
	meshes.insert(mesh);
	TraverseStack st;
	ElInfo *elInfo = st.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL);
	maxLevel_ = -1;
	while(elInfo) {
	  maxLevel_ = max(elInfo->getLevel(), maxLevel_);
	  elInfo = st.traverseNext(elInfo);   
	}
      }
    }
  }

  void MultiGridSolverVec::initMultiGrid(bool initMG)
  {
    int numLevels = maxLevel_ + 1;

    if (initMG) { 
      // Get number of components
      numComponents_ = solution_->getNumVectors();

      TEST_EXIT(rhs_->getNumVectors() == numComponents_)
	("invalid number of components\n");
      TEST_EXIT(systemMatrix_->getNumRows() == numComponents_)
	("invalid number of components\n");
      TEST_EXIT(systemMatrix_->getNumCols() == numComponents_)
	("invalid number of components\n");

      // Get fe spaces and create residual vectors
      sparseResidual_ = NEW Vector<SparseVector<double>*>(numComponents_);
      ::std::vector<FiniteElemSpace*> feSpaces(numComponents_);
      for (int i = 0; i < numComponents_; i++) {
	feSpaces[i] = const_cast<FiniteElemSpace*>
	  (solution_->getDOFVector(i)->getFESpace());
	(*sparseResidual_)[i] = NEW SparseVector<double>(feSpaces[i], 
							 "sparse residual");
      }    

      // Create dens residual vectors
      denseResidual_ = NEW SystemVector("dense residual", 
					feSpaces, 
					numComponents_);
      for (int i = 0; i < numComponents_; i++) {
	denseResidual_->setDOFVector(i, NEW DOFVector<double>(feSpaces[i],
							      "dense residual"));
      }

      Mesh *mesh = feSpaces[0]->getMesh();
      int dim = mesh->getDim();

      int error = mesh->queryCoarseDOFs();
      TEST_EXIT(error)("preserve coarse dofs not set in mesh\n");

      // Check which level is a multigrid level and set the corresponding field
      // to true.
      isMGLevel_.resize(numLevels, false);
      isMGLevel_[minLevel_] = true;
      isMGLevel_[maxLevel_] = true;
      if (numLevels > 2) { 
	double step = (static_cast<double>(maxMGLevels_) - 2.0) / (numLevels - 2.0);
	double sum = 0.0;
	int gap = 0;
	for (int i = maxLevel_ - 1; i > minLevel_; i--) {
	  sum += step;
	  //MSG("%d %f\n", i, sum);
	  ++gap;
	  if (gap > minLevelGap_ && sum >= 1.0) {
	    gap = 0;
	    isMGLevel_[i] = true;
	    sum -= 1.0;	
	  } else {
	    MSG("level %d skipped\n", i);
	  }
	}
      }

      levels_.resize(numLevels);

      // Create multigrid level data.
      for (int i = 0; i < numLevels; i++) {
	if (isMGLevel_[i]) {
	  levels_[i].components_.resize(numComponents_);
	  levels_[i].levelDOFs_.resize(numComponents_);

	  ::std::map<FiniteElemSpace*, FESpaceData*> feSpaceData;
	  MeshData *meshData = NEW MeshData;

	  levels_[i].solution_ = NEW Vector<SparseVector<double>*>(numComponents_);
	  levels_[i].oldSolution_ = NEW Vector<SparseVector<double>*>(numComponents_);
	  levels_[i].rhs_ = NEW Vector<SparseVector<double>*>(numComponents_);
	  levels_[i].matrix_ = NEW Matrix<DOFMatrix*>(numComponents_,
						      numComponents_);

	  // Set component dependend data.
	  for (int j = 0; j < numComponents_; j++) {
	    if (!feSpaceData[feSpaces[j]]) {
	      feSpaceData[feSpaces[j]] = NEW FESpaceData;
	    }
	    levels_[i].components_[j].meshData_ = meshData;
	    levels_[i].components_[j].feSpaceData_ = feSpaceData[feSpaces[j]];

	    levels_[i].levelDOFs_[j] = &(feSpaceData[feSpaces[j]]->levelDOFs_);

	    (*(levels_[i].solution_))[j] = 
	      NEW SparseVector<double>(feSpaces[j], "sparse level solution");
	    (*(levels_[i].oldSolution_))[j] = 
	      NEW SparseVector<double>("old level solution", 
				       (*(levels_[i].solution_))[j]);
	    (*(levels_[i].rhs_))[j] = 
	      NEW SparseVector<double>("sparse level rhs", 
				       (*(levels_[i].solution_))[j]);

	    for (int k = 0; k < numComponents_; k++) {
	      if ((*systemMatrix_)[j][k]) {
		(*(levels_[i].matrix_))[j][k] = NEW DOFMatrix(feSpaces[j], 
							      feSpaces[k],
							      "level matrix");
		(*(levels_[i].matrix_))[j][k]->
		  setBoundaryManager((*(systemMatrix_))[j][k]->
				     getBoundaryManager());
		if (j != k) {
		  (*(levels_[i].matrix_))[j][k]->setCoupleMatrix(true);
		}
	      } else {
		(*(levels_[i].matrix_))[j][k] = NULL;
	      }
	    }
	  }
	} // if isMGLevel_[i]
      } // for numLevels


      int degree;

      interpolMatrix_.resize(numComponents_);
      restrictMatrix_.resize(numComponents_);

      // Assemble matrices
      for (int row = 0; row < numComponents_; row++) {
	const BasisFunction* basFcts = feSpaces[row]->getBasisFcts();
	int numDOFs = basFcts->getNumber();

	degree = basFcts->getDegree();

	interpolMatrix_[row] = interpolMatrices[dim-1][degree-1];
	restrictMatrix_[row] = restrictMatrices[dim-1][degree-1];

	DegreeOfFreedom *dofs = GET_MEMORY(DegreeOfFreedom, numDOFs);
	Element *element;
	DOFAdmin *admin = feSpaces[row]->getAdmin();
	int level;

	TraverseStack stack;

	// fill max dof level for each dof
	DOFVector<int> maxDOFLevels(feSpaces[row], "maxDOFLevel");
	maxDOFLevels.set(minLevel_);
	ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_EVERY_EL_PREORDER);
	while (elInfo) {
	  element = elInfo->getElement();
	  level = elInfo->getLevel();

	  if (row == 0) {
	    levels_[level].elements_.push_back(element);
	  }
	  if (dim == 3) {
	    levels_[level].elementTypes_.
	      push_back(dynamic_cast<ElInfo3d*>(elInfo)->getType());
	  }


	  basFcts->getLocalIndices(element, admin, dofs);
	  for (int i = 0; i < numDOFs; i++) {
	    maxDOFLevels[dofs[i]] = max(maxDOFLevels[dofs[i]], 
					level);
	  }
	  elInfo = stack.traverseNext(elInfo);
	}

	for (int col = 0; col < numComponents_; col++) {
	  DOFMatrix *dofMatrix = (*systemMatrix_)[row][col];
	  if(dofMatrix) {
	    ::std::vector<Operator*>::iterator it;
	    ::std::vector<Operator*>::iterator opBegin = dofMatrix->getOperatorsBegin();
	    ::std::vector<Operator*>::iterator opEnd = dofMatrix->getOperatorsEnd();
	    ::std::vector<double*>::iterator factorIt;
	    ::std::vector<double*>::iterator factorBegin = dofMatrix->getOperatorFactorBegin();
	    ElementMatrix *elementMatrix = NULL; 
	    const BoundaryType *bound;
	
	    // assemble level matrices and create dof sets
	    elInfo = stack.traverseFirst(mesh, -1, 
					 Mesh::CALL_EVERY_EL_PREORDER |
					 Mesh::FILL_BOUND |
					 Mesh::FILL_DET |
					 Mesh::FILL_GRD_LAMBDA |
					 Mesh::FILL_COORDS);

	    while (elInfo) {
	      element = elInfo->getElement();
	      level = elInfo->getLevel();
	      basFcts->getLocalIndices(element, admin, dofs);
	      bound = basFcts->getBound(elInfo, NULL);
	      elementMatrix = (*opBegin)->getAssembler()->initElementMatrix(elementMatrix, elInfo);

	      bool levelUsed = isMGLevel_[level];
	      if (levelUsed) {
		for (it = opBegin, factorIt = factorBegin;	
		    it != opEnd; 
		    ++it, ++factorIt) 
		  {
		    (*it)->getElementMatrix(elInfo, 
					    elementMatrix, 
					    *factorIt ? **factorIt : 1.0);
		  }

		(*(levels_[level].matrix_))[row][col]->
		  addElementMatrix(1.0, *elementMatrix, bound);
	      }

	      if (element->isLeaf()) {
		if (!levelUsed) {
		  level = mgLevel(level);
		  for (it = opBegin, factorIt = factorBegin;	
		      it != opEnd; 
		      ++it, ++factorIt) 
		    {
		      (*it)->getElementMatrix(elInfo, 
					      elementMatrix, 
					      *factorIt ? **factorIt : 1.0);
		    }
		  //MSG("element %d added to level %d\n", element->getIndex(), level);
		  (*(levels_[level].matrix_))[row][col]->
		    addElementMatrix(1.0, *elementMatrix, bound);
		} 

		// add nearest neighbour element matrices and dofs
		int maxDOFLevel = -1;
		for (int i = 0; i < dim + 1; i++) {
		  maxDOFLevel = max(maxDOFLevel, mgLevel(maxDOFLevels[dofs[i]]));
		}

		if (maxDOFLevel > level) {
		  for (int j = level + 1; j < maxDOFLevel + 1; j++) {
		    if (isMGLevel_[j]) {
		      (*(levels_[j].matrix_))[row][col]->
			addElementMatrix(1.0, *elementMatrix, bound);
		    }
		  }
		  if (row == col) {
		    for (int j = 0; j < numDOFs; j++) {
		      for (int k = level + 1; k < maxDOFLevel + 1; k++) {
			if (isMGLevel_[k]) {
			  if (mgLevel(maxDOFLevels[dofs[j]]) < k) {
			    levels_[k].components_[row].feSpaceData_->
			      nearestNeighbourDOFs_.insert(dofs[j]);
			    //MSG("level %d comp %d nn: %d\n", k, row, dofs[j]);
			  }
			}
		      }
		    }
		  }
		}
		if (row == col) {
		  for (int i = 0; i < numDOFs; i++) {
		    if (mgLevel(maxDOFLevels[dofs[i]]) == level) {
		      levels_[level].components_[row].feSpaceData_->
			coarseDOFs_.insert(dofs[i]);
		      levels_[level].components_[row].feSpaceData_->
			levelDOFs_.insert(dofs[i]);
		      //MSG("level %d comp %d coarse: %d\n", level, row, dofs[i]);
		    }
		  }
		}
	      } else { // element is not leaf
		if (row == col) {
		  if (level >= minLevel_) {
		    if (levelUsed) {
		      for (int i = 0; i < numDOFs; i++) {
			levels_[level].components_[row].feSpaceData_->
			  fineDOFs_.insert(dofs[i]);
			levels_[level].components_[row].feSpaceData_->
			  levelDOFs_.insert(dofs[i]);
			//MSG("level %d comp %d fine: %d\n", level, row, dofs[i]);
		      }
		    }
		  }
		}
	      }
	      elInfo = stack.traverseNext(elInfo);
	    }
	  }
	}

	FREE_MEMORY(dofs, DegreeOfFreedom, numDOFs);
      }

      if (galerkin_) {
	int level;
	TraverseStack stack;
	ElInfo *elInfo;
	const FiniteElemSpace *feSpace = feSpaces[0];
	const BasisFunction *basFcts = feSpace->getBasisFcts();
	DOFAdmin *admin = feSpace->getAdmin();
	Mesh *mesh = feSpace->getMesh();

	for (int row = 0; row < numComponents_; row++) {
	  for (int col = 0; col < numComponents_; col++) {
	    if ((*(levels_[maxLevel_].matrix_))[row][col]) {
	      (*(levels_[maxLevel_].matrix_))[row][col]->copy(*((*systemMatrix_)[row][col]));
	    }
	  }
	}

	DegreeOfFreedom *child0DOFs = GET_MEMORY(DegreeOfFreedom, dim + 1);
	DegreeOfFreedom *parentDOFs = GET_MEMORY(DegreeOfFreedom, dim + 1);

	Vector<DegreeOfFreedom> fineDOFs(dim + 2);
	Matrix<double> a(dim + 2, dim + 2);

	DOFMatrix tmp1(feSpace, feSpace, "tmp1");
	DOFMatrix tmp2(feSpace, feSpace, "tmp2");

	DOFMatrix *coarseMatrix = NULL;
	DOFMatrix *fineMatrix = NULL;

	for (int row = 0; row < numComponents_; row++) {
	  for (int col = 0; col < numComponents_; col++) {
	    if ((*systemMatrix_)[row][col]) {
	      fineMatrix = (*(levels_[maxLevel_].matrix_))[row][col];

	      for (level = maxLevel_ - 1; level >= 0; level--) {
		if (level >= minLevel_) {

		  ::std::map< ::std::pair<int, int>, bool> visited;
		  coarseMatrix = 
		    isMGLevel_[level] ?
		    (*(levels_[level].matrix_))[row][col] :
		    (fineMatrix == &tmp1 ? &tmp2 : &tmp1);

		  coarseMatrix->clear();

		  elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_EVERY_EL_PREORDER);
		  while (elInfo) {
		    Element *element = elInfo->getElement();

		    basFcts->getLocalIndices(element, admin, parentDOFs);
	      
		    for (int i = 0; i < dim + 1; i++) {
		      fineDOFs[i] = parentDOFs[i];
		    }

		    if (elInfo->getLevel() == level && !element->isLeaf()) {
		      basFcts->getLocalIndices(element->getChild(0), admin, child0DOFs);

		      fineDOFs[dim + 1] = child0DOFs[dim];

		      double *ptr;

		      for (int i = 0; i < dim + 2; i++) {
			for (int j = 0; j < dim + 2; j++) {
			  ptr = fineMatrix->hasSparseDOFEntry(fineDOFs[i], fineDOFs[j]);
			  a[i][j] = ptr ? *ptr : 0.0;
			}
		      }

		      for (int i = 0; i < dim + 1; i++) {
			for(int j = 0; j < dim + 1; j++) {
			  if (visited[::std::pair<int, int>(fineDOFs[i], fineDOFs[j])] == false) {
			    coarseMatrix->addSparseDOFEntry(1.0, 
							    parentDOFs[i],
							    parentDOFs[j],
							    0.0,
							    false);		
			  }
			}
		      }
		  
		      for (int i = 0; i < dim + 1; i++) {
			for (int j = 0; j < dim + 1; j++) {
			  if (visited[::std::pair<int, int>(fineDOFs[i], fineDOFs[j])] == false) {
			    coarseMatrix->addSparseDOFEntry(1.0, 
							    parentDOFs[i],
							    parentDOFs[j],
							    a[i][j],
							    true);
			  }
			}
		      }
		  
		      for (int i = 0; i < dim + 1; i++) {
			for (int j = 0; j < 2; j++) {
			  if (visited[::std::pair<int, int>(fineDOFs[i], fineDOFs[dim + 1])] == false) {
			    coarseMatrix->addSparseDOFEntry(0.5, 
							    parentDOFs[i],
							    parentDOFs[j],
							    a[i][dim + 1],
							    true);	      
			
			    coarseMatrix->addSparseDOFEntry(0.5, 
							    parentDOFs[j],
							    parentDOFs[i],
							    a[dim + 1][i],
							    true);
			  }	      
			}
		      }
		  
		      if (visited[::std::pair<int, int>(fineDOFs[dim + 1], fineDOFs[dim + 1])] == false) {
			for (int i = 0; i < 2; i++) {
			  for (int j = 0; j < 2; j++) {
			    coarseMatrix->addSparseDOFEntry(0.25, 
							    parentDOFs[i],
							    parentDOFs[j],
							    a[dim + 1][dim + 1],
							    true);	      
			  }
			}
		      }
		
		      for (int i = 0; i < dim + 2; i++) {
			for (int j = 0; j < dim + 2; j++) {
			  visited[::std::pair<int, int>(fineDOFs[i], fineDOFs[j])] = true;
			  visited[::std::pair<int, int>(fineDOFs[j], fineDOFs[i])] = true;
			}
		      }
		    } 

		    if (element->isLeaf() && elInfo->getLevel() <= level) {
		      double *ptr;
		  
		      for (int i = 0; i < dim + 1; i++) {
			for (int j = 0; j < dim + 1; j++) {
			  ptr = fineMatrix->hasSparseDOFEntry(fineDOFs[i], fineDOFs[j]);
			  a[i][j] = ptr ? *ptr : 0.0;
			}
		      }
		  
		      for (int i = 0; i < dim + 1; i++) {
			for (int j = 0; j < dim + 1; j++) {
			  if (!visited[::std::pair<int, int>(parentDOFs[i], parentDOFs[j])]) {
			    coarseMatrix->addSparseDOFEntry(1.0, 
							    parentDOFs[i],
							    parentDOFs[j],
							    a[i][j],
							    false);	      		
			  }
			}
		      }	  
		    }
		    elInfo = stack.traverseNext(elInfo);
		  }
		  fineMatrix = coarseMatrix;
		}
	      }
	    }
	  }
	}
      
	FREE_MEMORY(child0DOFs, DegreeOfFreedom, dim + 1);
	FREE_MEMORY(parentDOFs, DegreeOfFreedom, dim + 1);
      }

      if (boxSmoothing_) {
	boxDOFs_.resize(numComponents_);
	for (int i = 0; i < numComponents_; i++) {
	  boxDOFs_[i] = new ::std::set<DegreeOfFreedom>;
	}
      }

    }


    iteration_ = 0;

    lastRes_ = 1.0;

    // compute start residual
    mv(*systemMatrix_, *solution_, *denseResidual_);
    axpy(-1.0, *rhs_, *denseResidual_);

    double res = norm(denseResidual_);
    MSG("residual: %e\n", res);
    lastRes_ = res;

    for (int row = 0; row < numComponents_; row++) {
      for (int i = 0; i < numLevels; i++) {
	if (isMGLevel_[i]) {
	  (*(levels_[i].solution_))[row]->
	    copyFromDOFVector(*(solution_->getDOFVector(row)), 
			      levels_[i].components_[row].feSpaceData_->coarseDOFs_);
	  (*(levels_[i].solution_))[row]->
	    copyFromDOFVector(*(solution_->getDOFVector(row)), 
			      levels_[i].components_[row].feSpaceData_->nearestNeighbourDOFs_);
	}
      }
  

      (*(levels_[maxLevel_].rhs_))[row]->
	copyFromDOFVector(*(rhs_->getDOFVector(row)), 
			  levels_[maxLevel_].components_[row].feSpaceData_->coarseDOFs_);
    }
  }


  void MultiGridSolverVec::exitMultiGrid()
  {
    int numLevels = maxLevel_ + 1;

    ::std::set<FESpaceData*> deletedFESpaceData;
    ::std::set<MeshData*> deletedMeshData;

    for (int i = 0; i < numLevels; i++) {
      if (isMGLevel_[i]) {
	for (int j = 0; j < numComponents_; j++) {
	  if (levels_[i].components_[j].meshData_) {
	    ::std::set<MeshData*>::iterator it;
	    it = ::std::find(deletedMeshData.begin(), 
			     deletedMeshData.end(),
			     levels_[i].components_[j].meshData_);
	    if (it == deletedMeshData.end()) {
	      deletedMeshData.insert(levels_[i].components_[j].meshData_);
	      DELETE levels_[i].components_[j].meshData_;
	      levels_[i].components_[j].meshData_ = NULL;
	    }
	  }
	  if (levels_[i].components_[j].feSpaceData_) {
	    ::std::set<FESpaceData*>::iterator it;
	    it = ::std::find(deletedFESpaceData.begin(), 
			     deletedFESpaceData.end(),
			     levels_[i].components_[j].feSpaceData_);
	    if (it == deletedFESpaceData.end()) {
	      deletedFESpaceData.insert(levels_[i].components_[j].feSpaceData_);
	      DELETE levels_[i].components_[j].feSpaceData_;
	      levels_[i].components_[j].feSpaceData_ = NULL;
	    }
	  }
	  DELETE (*(levels_[i].oldSolution_))[j];
	  DELETE (*(levels_[i].rhs_))[j];
	  DELETE (*(levels_[i].solution_))[j];

	  for (int k = 0; k < numComponents_; k++) {
	    if ((*(levels_[i].matrix_))[j][k]) {
	      DELETE (*(levels_[i].matrix_))[j][k];
	    }
	  }
	}
	DELETE levels_[i].oldSolution_;
	DELETE levels_[i].solution_;
	DELETE levels_[i].rhs_;
	DELETE levels_[i].matrix_;
      }
    }
    levels_.resize(0);
    isMGLevel_.resize(0);

    if (boxSmoothing_) {
      for (int i = 0; i < numComponents_; i++) {
	delete boxDOFs_[i];
      }
      boxDOFs_.resize(0);
    }

    for (int i = 0; i < numComponents_; i++) {
      DELETE (*sparseResidual_)[i];
      DELETE denseResidual_->getDOFVector(i);
    }
    DELETE sparseResidual_;
    DELETE denseResidual_;

    numComponents_ = 0;
  }


  void MultiGridSolverVec::smooth(int level, int steps)
  { 
    if (boxSmoothing_) {
      for (int iteration = 0; iteration < steps; iteration++) {    
	::std::vector<Element*>::iterator elIt, elBegin, elEnd;
	elBegin = levels_[level].elements_.begin();
	elEnd   = levels_[level].elements_.end();
	for (elIt = elBegin; elIt != elEnd; ++elIt) {
	  Element *element = *elIt;
	  for (int i = 0; i < numComponents_; i++) {
	    boxDOFs_[i]->clear();
	    const FiniteElemSpace *feSpace = 
	      (*(levels_[level].solution_))[i]->getFESpace();
	    int numDOFs = feSpace->getBasisFcts()->getNumber();
	    DOFAdmin *admin = feSpace->getAdmin();
	    DegreeOfFreedom *dofs = GET_MEMORY(DegreeOfFreedom, numDOFs);
	    feSpace->getBasisFcts()->getLocalIndices(element,
						     admin,
						     dofs);
	    for (int j = 0; j < numDOFs; j++) {
	      boxDOFs_[i]->insert(dofs[j]);
	    }
	    FREE_MEMORY(dofs, DegreeOfFreedom, numDOFs);
	  }

	  smoother_->smooth(levels_[level].matrix_,
			    levels_[level].solution_,
			    levels_[level].rhs_,
			    1,
			    boxDOFs_);
	}
      }
    } else {
      smoother_->smooth(levels_[level].matrix_,
			levels_[level].solution_,
			levels_[level].rhs_,
			steps,
			levels_[level].levelDOFs_);

    }

    //   DOFVector<double> sol0((*(systemMatrix_))[0][0]->getFESpace(), "sol0");
    //   (*(levels_[level].solution_))[0]->copyToDOFVector(sol0);
    //   MacroWriter::writeMacro((*(systemMatrix_))[0][0]->getFESpace(), "sol0.mesh", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&sol0, "sol0.dat", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   DOFVector<double> sol1((*(systemMatrix_))[1][1]->getFESpace(), "sol1");
    //   (*(levels_[level].solution_))[1]->copyToDOFVector(sol1);
    //   MacroWriter::writeMacro((*(systemMatrix_))[1][1]->getFESpace(), "sol1.mesh", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   ValueWriter::writeValues(&sol1, "sol1.dat", 0.0, level, Mesh::CALL_EL_LEVEL);
    //   MSG("smoothed %d\n", level);
    //   WAIT; 
  }


  void MultiGridSolverVec::prolongate(int level, int coarseLevel)
  {
    for (int comp = 0; comp < numComponents_; comp++) {
      for (int i = coarseLevel; i < level; i++) {
	interpolMatrix_[comp]->mv((*sparseResidual_)[comp],
				  (*sparseResidual_)[comp],
				  i);
      }
    }
  }


  void MultiGridSolverVec::restrict(int level, int coarseLevel)
  {
    ::std::vector<Element*>::iterator elIt, elBegin, elEnd;
    ::std::vector<int>::iterator typeIt;

    for (int comp = 0; comp < numComponents_; comp++) {
      (*(levels_[coarseLevel].oldSolution_))[comp]->
	copy(*((*(levels_[level].solution_))[comp]),
	     levels_[level].components_[comp].feSpaceData_->levelDOFs_);
      (*(levels_[coarseLevel].oldSolution_))[comp]->
	copy(*((*(levels_[level].solution_))[comp]),
	     levels_[level].components_[comp].feSpaceData_->nearestNeighbourDOFs_);

      for (int i = level; i > coarseLevel; i--) {
	SparseVector<double> visited("visited", (*sparseResidual_)[comp]);

	elBegin = levels_[i-1].elements_.begin();
	elEnd = levels_[i-1].elements_.end();

	typeIt = levels_[i-1].elementTypes_.begin();
	for (elIt = elBegin; elIt != elEnd; ++elIt) {
	  Element *element = *elIt;
	  if (!element->isLeaf()) {
	    RCNeighbourList list(1);
	    list.setElement(0, element);
	    if (element->getMesh()->getDim() == 3) {
	      list.setType(0, *typeIt);
	    }
	    (*(levels_[coarseLevel].oldSolution_))[comp]->coarseRestrict(list, 1);
	    //sparseResidual_->coarseRestrict(list, 1);
	  }
	}

	//       restrictMatrix_[comp]->mv((*(levels_[coarseLevel].oldSolution_))[comp],
	// 				(*(levels_[coarseLevel].oldSolution_))[comp],
	// 				i-1);
	restrictMatrix_[comp]->mv((*sparseResidual_)[comp],
				  (*sparseResidual_)[comp],
				  i-1);
      }

      (*(levels_[coarseLevel].solution_))[comp]->
	copy(*((*(levels_[coarseLevel].oldSolution_))[comp]),
	     levels_[coarseLevel].components_[comp].feSpaceData_->fineDOFs_);
    }
  }


  void MultiGridSolverVec::computeResidual(int level, int coarseLevel)
  {
    DOFMatrix *matrix;
    SparseVector<double> *res, *sol, *rhs;
    ::std::set<DegreeOfFreedom> *dofSet;
    for (int i = 0; i < numComponents_; i++) {
      res = (*sparseResidual_)[i];
      res->clear();
      rhs = (*(levels_[level].rhs_))[i];
      dofSet = levels_[level].levelDOFs_[i];
      for (int j = 0; j < numComponents_; j++) {
	matrix = (*(levels_[level].matrix_))[i][j];
	if (matrix) {
	  sol = (*(levels_[level].solution_))[j];
	  mv(*matrix, *sol, *res, *dofSet, true);
	}
      }
      axpy(-1.0, *rhs, *res, *dofSet);
    }
  }


  void MultiGridSolverVec::computeRHS(int level, int coarseLevel)
  {
    DOFMatrix *matrix;
    SparseVector<double> *oldSol, *rhs, *res;
    ::std::set<DegreeOfFreedom> *dofSet;
    for (int i = 0; i < numComponents_; i++) {
      rhs = (*(levels_[coarseLevel].rhs_))[i];
      rhs->clear();
      res = (*sparseResidual_)[i];
      dofSet = &(levels_[coarseLevel].components_[i].feSpaceData_->fineDOFs_);
      for (int j = 0; j < numComponents_; j++) {
	matrix = (*(levels_[coarseLevel].matrix_))[i][j];
	if (matrix) {
	  oldSol = (*(levels_[coarseLevel].oldSolution_))[j];
	  mv(*matrix, *oldSol, *rhs, *dofSet, true);
	}
      }
      axpy(-1.0, *res, *rhs, *dofSet);
      dofSet = &(levels_[coarseLevel].components_[i].feSpaceData_->coarseDOFs_);
      rhs->copyFromDOFVector(*(rhs_->getDOFVector(i)), *dofSet);
    }
  }


  void MultiGridSolverVec::computeCorrection(int level, int coarseLevel)
  {
    SparseVector<double> *res, *sol, *oldSol;
    ::std::set<DegreeOfFreedom> *dofSet;
    for (int comp = 0; comp < numComponents_; comp++) {
      dofSet = &(levels_[coarseLevel].components_[comp].feSpaceData_->fineDOFs_);
      sol = (*(levels_[coarseLevel].solution_))[comp];
      oldSol = (*(levels_[coarseLevel].oldSolution_))[comp];
      res = (*sparseResidual_)[comp];
      res->copy(*sol, *dofSet);
      axpy(-1.0, *oldSol, *res, *dofSet); 
    }
  } 


  void MultiGridSolverVec::fixSolution(int level, int coarseLevel)
  {
    SparseVector<double> *fineSol, *coarseSol, *res;
    ::std::set<DegreeOfFreedom> *dofSet;
    for (int comp = 0; comp < numComponents_; comp++) {
      fineSol = (*(levels_[level].solution_))[comp];
      coarseSol = (*(levels_[coarseLevel].solution_))[comp];
      res = (*sparseResidual_)[comp];
      dofSet = 
	&(levels_[level].components_[comp].feSpaceData_->nearestNeighbourDOFs_);
      fineSol->copy(*coarseSol, *dofSet);
      dofSet = 
	&(levels_[level].components_[comp].feSpaceData_->levelDOFs_);
      axpy(1.0, *res, *fineSol, *dofSet);
    }
  }


  void MultiGridSolverVec::exactSolve(int level)
  {
    smooth(level, coarseSmoothingSteps_);
  }


  bool MultiGridSolverVec::toleranceReached(double tol)
  { 
    FUNCNAME("MultiGridSolverVec::toleranceReached()");

    SparseVector<double> *sol;
    ::std::set<DegreeOfFreedom> *dofSet;
    for (int level = minLevel_; level < maxLevel_ + 1; level++) {
      if (isMGLevel_[level]) {
	for (int i = 0; i < numComponents_; i++) {
	  sol = (*(levels_[level].solution_))[i];
	  dofSet = &(levels_[level].components_[i].feSpaceData_->coarseDOFs_);
	  sol->copyToDOFVector(*(solution_->getDOFVector(i)), *dofSet);
	}
      }
    }

    mv(*systemMatrix_, *solution_, *denseResidual_);
    axpy(-1.0, *rhs_, *denseResidual_);
  
    ++iteration_;

    double res = norm(denseResidual_);

    MSG("iteration %d - residual: %e (res/lastRes: %e)\n", 
	iteration_, res, res/lastRes_);
    lastRes_ = res;

    return (res < tol);
  }


  void MultiGridSolverVec::initFullMultiGrid()
  {
    FUNCNAME("MultiGridSolverVec::initFullMultiGrid()");
    ERROR_EXIT("not yet\n");
  }


  void MultiGridSolverVec::fmgInterpol(int level)
  {
    FUNCNAME("MultiGridSolverVec::fmgInterpol()");
    ERROR_EXIT("not yet\n");
  }


}
